@@ -1,4 +1,4 @@
-# Copyright 2009-2015 MongoDB, Inc.
+# Copyright 2009-present MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -125,7 +125,7 @@ class GridFS(object):
 
         return grid_file._id
 
-    def get(self, file_id):
+    def get(self, file_id, session=None):
         """Get a file from GridFS by ``"_id"``.
 
         Returns an instance of :class:`~gridfs.grid_file.GridOut`,
@@ -133,14 +133,19 @@ class GridFS(object):
 
         :Parameters:
           - `file_id`: ``"_id"`` of the file to get
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
         """
-        gout = GridOut(self.__collection, file_id)
+        gout = GridOut(self.__collection, file_id, session=session)
 
         # Raise NoFile now, instead of on first attribute access.
         gout._ensure_file()
         return gout
 
-    def get_version(self, filename=None, version=-1, **kwargs):
+    def get_version(self, filename=None, version=-1, session=None, **kwargs):
         """Get a file from GridFS by ``"filename"`` or metadata fields.
 
         Returns a version of the file in GridFS whose filename matches
@@ -165,8 +170,13 @@ class GridFS(object):
           - `filename`: ``"filename"`` of the file to get, or `None`
           - `version` (optional): version of the file to get (defaults
             to -1, the most recent version uploaded)
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
           - `**kwargs` (optional): find files by custom metadata.
 
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
+
         .. versionchanged:: 3.1
            ``get_version`` no longer ensures indexes.
         """
@@ -174,19 +184,20 @@ class GridFS(object):
         if filename is not None:
             query["filename"] = filename
 
-        cursor = self.__files.find(query)
+        cursor = self.__files.find(query, session=session)
         if version < 0:
             skip = abs(version) - 1
             cursor.limit(-1).skip(skip).sort("uploadDate", DESCENDING)
         else:
             cursor.limit(-1).skip(version).sort("uploadDate", ASCENDING)
         try:
-            grid_file = next(cursor)
-            return GridOut(self.__collection, file_document=grid_file)
+            doc = next(cursor)
+            return GridOut(
+                self.__collection, file_document=doc, session=session)
         except StopIteration:
             raise NoFile("no version %d for filename %r" % (version, filename))
 
-    def get_last_version(self, filename=None, **kwargs):
+    def get_last_version(self, filename=None, session=None, **kwargs):
         """Get the most recent version of a file in GridFS by ``"filename"``
         or metadata fields.
 
@@ -195,12 +206,17 @@ class GridFS(object):
 
         :Parameters:
           - `filename`: ``"filename"`` of the file to get, or `None`
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
           - `**kwargs` (optional): find files by custom metadata.
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
         """
-        return self.get_version(filename=filename, **kwargs)
+        return self.get_version(filename=filename, session=session, **kwargs)
 
     # TODO add optional safe mode for chunk removal?
-    def delete(self, file_id):
+    def delete(self, file_id, session=None):
         """Delete a file from GridFS by ``"_id"``.
 
         Deletes all data belonging to the file with ``"_id"``:
@@ -216,27 +232,39 @@ class GridFS(object):
 
         :Parameters:
           - `file_id`: ``"_id"`` of the file to delete
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
 
         .. versionchanged:: 3.1
            ``delete`` no longer ensures indexes.
         """
-        self.__files.delete_one({"_id": file_id})
-        self.__chunks.delete_many({"files_id": file_id})
+        self.__files.delete_one({"_id": file_id}, session=session)
+        self.__chunks.delete_many({"files_id": file_id}, session=session)
 
-    def list(self):
+    def list(self, session=None):
         """List the names of all files stored in this instance of
         :class:`GridFS`.
 
+        :Parameters:
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
+
         .. versionchanged:: 3.1
            ``list`` no longer ensures indexes.
         """
         # With an index, distinct includes documents with no filename
         # as None.
         return [
-            name for name in self.__files.distinct("filename")
+            name for name in self.__files.distinct("filename", session=session)
             if name is not None]
 
-    def find_one(self, filter=None, *args, **kwargs):
+    def find_one(self, filter=None, session=None, *args, **kwargs):
         """Get a single file from gridfs.
 
         All arguments to :meth:`find` are also valid arguments for
@@ -252,13 +280,18 @@ class GridFS(object):
             the value for a query for ``"_id"`` in the file collection.
           - `*args` (optional): any additional positional arguments are
             the same as the arguments to :meth:`find`.
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
           - `**kwargs` (optional): any additional keyword arguments
             are the same as the arguments to :meth:`find`.
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
         """
         if filter is not None and not isinstance(filter, Mapping):
             filter = {"_id": filter}
 
-        for f in self.find(filter, *args, **kwargs):
+        for f in self.find(filter, *args, session=session, **kwargs):
             return f
 
         return None
@@ -290,6 +323,10 @@ class GridFS(object):
         :meth:`~pymongo.collection.Collection.find`
         in :class:`~pymongo.collection.Collection`.
 
+        If a :class:`~pymongo.client_session.ClientSession` is passed to
+        :meth:`find`, all returned :class:`~gridfs.grid_file.GridOut` instances
+        are associated with that session.
+
         :Parameters:
           - `filter` (optional): a SON object specifying elements which
             must be present for a document to be included in the
@@ -320,7 +357,7 @@ class GridFS(object):
         """
         return GridOutCursor(self.__collection, *args, **kwargs)
 
-    def exists(self, document_or_id=None, **kwargs):
+    def exists(self, document_or_id=None, session=None, **kwargs):
         """Check if a file exists in this instance of :class:`GridFS`.
 
         The file to check for can be specified by the value of its
@@ -350,12 +387,20 @@ class GridFS(object):
         :Parameters:
           - `document_or_id` (optional): query document, or _id of the
             document to check for
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
           - `**kwargs` (optional): keyword arguments are used as a
             query document, if they're present.
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
         """
         if kwargs:
-            return self.__files.find_one(kwargs, ["_id"]) is not None
-        return self.__files.find_one(document_or_id, ["_id"]) is not None
+            f = self.__files.find_one(kwargs, ["_id"], session=session)
+        else:
+            f = self.__files.find_one(document_or_id, ["_id"], session=session)
+
+        return f is not None
 
 
 class GridFSBucket(object):
@@ -409,7 +454,7 @@ class GridFSBucket(object):
         self._chunk_size_bytes = chunk_size_bytes
 
     def open_upload_stream(self, filename, chunk_size_bytes=None,
-                           metadata=None):
+                           metadata=None, session=None):
         """Opens a Stream that the application can write the contents of the
         file to.
 
@@ -439,6 +484,11 @@ class GridFSBucket(object):
           - `metadata` (optional): User data for the 'metadata' field of the
             files collection document. If not provided the metadata field will
             be omitted from the files collection document.
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
         """
         validate_string("filename", filename)
 
@@ -448,10 +498,63 @@ class GridFSBucket(object):
         if metadata is not None:
             opts["metadata"] = metadata
 
-        return GridIn(self._collection, **opts)
+        return GridIn(self._collection, session=session, **opts)
+
+    def open_upload_stream_with_id(
+            self, file_id, filename, chunk_size_bytes=None, metadata=None,
+            session=None):
+        """Opens a Stream that the application can write the contents of the
+        file to.
+
+        The user must specify the file id and filename, and can choose to add
+        any additional information in the metadata field of the file document
+        or modify the chunk size.
+        For example::
+
+          my_db = MongoClient().test
+          fs = GridFSBucket(my_db)
+          grid_in, file_id = fs.open_upload_stream(
+                ObjectId(),
+                "test_file",
+                chunk_size_bytes=4,
+                metadata={"contentType": "text/plain"})
+          grid_in.write("data I want to store!")
+          grid_in.close()  # uploaded on close
+
+        Returns an instance of :class:`~gridfs.grid_file.GridIn`.
+
+        Raises :exc:`~gridfs.errors.NoFile` if no such version of
+        that file exists.
+        Raises :exc:`~ValueError` if `filename` is not a string.
+
+        :Parameters:
+          - `file_id`: The id to use for this file. The id must not have
+            already been used for another file.
+          - `filename`: The name of the file to upload.
+          - `chunk_size_bytes` (options): The number of bytes per chunk of this
+            file. Defaults to the chunk_size_bytes in :class:`GridFSBucket`.
+          - `metadata` (optional): User data for the 'metadata' field of the
+            files collection document. If not provided the metadata field will
+            be omitted from the files collection document.
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
+        """
+        validate_string("filename", filename)
+
+        opts = {"_id": file_id,
+                "filename": filename,
+                "chunk_size": (chunk_size_bytes if chunk_size_bytes
+                               is not None else self._chunk_size_bytes)}
+        if metadata is not None:
+            opts["metadata"] = metadata
+
+        return GridIn(self._collection, session=session, **opts)
 
     def upload_from_stream(self, filename, source, chunk_size_bytes=None,
-                           metadata=None):
+                           metadata=None, session=None):
         """Uploads a user file to a GridFS bucket.
 
         Reads the contents of the user file from `source` and uploads
@@ -481,14 +584,63 @@ class GridFSBucket(object):
           - `metadata` (optional): User data for the 'metadata' field of the
             files collection document. If not provided the metadata field will
             be omitted from the files collection document.
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
         """
         with self.open_upload_stream(
-                filename, chunk_size_bytes, metadata) as gin:
+                filename, chunk_size_bytes, metadata, session=session) as gin:
             gin.write(source)
 
         return gin._id
 
-    def open_download_stream(self, file_id):
+    def upload_from_stream_with_id(self, file_id, filename, source,
+                                   chunk_size_bytes=None, metadata=None,
+                                   session=None):
+        """Uploads a user file to a GridFS bucket with a custom file id.
+
+        Reads the contents of the user file from `source` and uploads
+        it to the file `filename`. Source can be a string or file-like object.
+        For example::
+
+          my_db = MongoClient().test
+          fs = GridFSBucket(my_db)
+          file_id = fs.upload_from_stream(
+              ObjectId(),
+              "test_file",
+              "data I want to store!",
+              chunk_size_bytes=4,
+              metadata={"contentType": "text/plain"})
+
+        Raises :exc:`~gridfs.errors.NoFile` if no such version of
+        that file exists.
+        Raises :exc:`~ValueError` if `filename` is not a string.
+
+        :Parameters:
+          - `file_id`: The id to use for this file. The id must not have
+            already been used for another file.
+          - `filename`: The name of the file to upload.
+          - `source`: The source stream of the content to be uploaded. Must be
+            a file-like object that implements :meth:`read` or a string.
+          - `chunk_size_bytes` (options): The number of bytes per chunk of this
+            file. Defaults to the chunk_size_bytes of :class:`GridFSBucket`.
+          - `metadata` (optional): User data for the 'metadata' field of the
+            files collection document. If not provided the metadata field will
+            be omitted from the files collection document.
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
+        """
+        with self.open_upload_stream_with_id(
+                file_id, filename, chunk_size_bytes, metadata,
+                session=session) as gin:
+            gin.write(source)
+
+    def open_download_stream(self, file_id, session=None):
         """Opens a Stream from which the application can read the contents of
         the stored file specified by file_id.
 
@@ -506,15 +658,20 @@ class GridFSBucket(object):
         Raises :exc:`~gridfs.errors.NoFile` if no file with file_id exists.
 
         :Parameters:
-          -`file_id`: The _id of the file to be downloaded.
+          - `file_id`: The _id of the file to be downloaded.
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
         """
-        gout = GridOut(self._collection, file_id)
+        gout = GridOut(self._collection, file_id, session=session)
 
         # Raise NoFile now, instead of on first attribute access.
         gout._ensure_file()
         return gout
 
-    def download_to_stream(self, file_id, destination):
+    def download_to_stream(self, file_id, destination, session=None):
         """Downloads the contents of the stored file specified by file_id and
         writes the contents to `destination`.
 
@@ -525,20 +682,27 @@ class GridFSBucket(object):
           # Get _id of file to read
           file_id = fs.upload_from_stream("test_file", "data I want to store!")
           # Get file to write to
-          file = open('myfile','rw')
+          file = open('myfile','wb+')
           fs.download_to_stream(file_id, file)
+          file.seek(0)
           contents = file.read()
 
         Raises :exc:`~gridfs.errors.NoFile` if no file with file_id exists.
 
         :Parameters:
-          -`file_id`: The _id of the file to be downloaded.
-          -`destination`: a file-like object implementing :meth:`write`.
-        """
-        gout = self.open_download_stream(file_id)
-        destination.write(gout)
+          - `file_id`: The _id of the file to be downloaded.
+          - `destination`: a file-like object implementing :meth:`write`.
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
 
-    def delete(self, file_id):
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
+        """
+        gout = self.open_download_stream(file_id, session=session)
+        for chunk in gout:
+            destination.write(chunk)
+
+    def delete(self, file_id, session=None):
         """Given an file_id, delete this stored file's files collection document
         and associated chunks from a GridFS bucket.
 
@@ -547,16 +711,21 @@ class GridFSBucket(object):
           my_db = MongoClient().test
           fs = GridFSBucket(my_db)
           # Get _id of file to delete
-         file_id = fs.upload_from_stream("test_file", "data I want to store!")
+          file_id = fs.upload_from_stream("test_file", "data I want to store!")
           fs.delete(file_id)
 
         Raises :exc:`~gridfs.errors.NoFile` if no file with file_id exists.
 
         :Parameters:
-          -`file_id`: The _id of the file to be deleted.
+          - `file_id`: The _id of the file to be deleted.
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
         """
-        res = self._files.delete_one({"_id": file_id})
-        self._chunks.delete_many({"files_id": file_id})
+        res = self._files.delete_one({"_id": file_id}, session=session)
+        self._chunks.delete_many({"files_id": file_id}, session=session)
         if not res.deleted_count:
             raise NoFile(
                 "no file could be deleted because none matched %s" % file_id)
@@ -590,6 +759,10 @@ class GridFSBucket(object):
         :meth:`~pymongo.collection.Collection.find`
         in :class:`~pymongo.collection.Collection`.
 
+        If a :class:`~pymongo.client_session.ClientSession` is passed to
+        :meth:`find`, all returned :class:`~gridfs.grid_file.GridOut` instances
+        are associated with that session.
+
         :Parameters:
           - `filter`: Search query.
           - `batch_size` (optional): The number of documents to return per
@@ -605,7 +778,7 @@ class GridFSBucket(object):
         """
         return GridOutCursor(self._collection, *args, **kwargs)
 
-    def open_download_stream_by_name(self, filename, revision=-1):
+    def open_download_stream_by_name(self, filename, revision=-1, session=None):
         """Opens a Stream from which the application can read the contents of
         `filename` and optional `revision`.
 
@@ -628,20 +801,26 @@ class GridFSBucket(object):
           - `revision` (optional): Which revision (documents with the same
             filename and different uploadDate) of the file to retrieve.
             Defaults to -1 (the most recent revision).
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
 
         :Note: Revision numbers are defined as follows:
-          0 = the original stored file
-          1 = the first revision
-          2 = the second revision
-          etc...
-          -2 = the second most recent revision
-          -1 = the most recent revision
+
+          - 0 = the original stored file
+          - 1 = the first revision
+          - 2 = the second revision
+          - etc...
+          - -2 = the second most recent revision
+          - -1 = the most recent revision
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
         """
         validate_string("filename", filename)
 
         query = {"filename": filename}
 
-        cursor = self._files.find(query)
+        cursor = self._files.find(query, session=session)
         if revision < 0:
             skip = abs(revision) - 1
             cursor.limit(-1).skip(skip).sort("uploadDate", DESCENDING)
@@ -649,12 +828,14 @@ class GridFSBucket(object):
             cursor.limit(-1).skip(revision).sort("uploadDate", ASCENDING)
         try:
             grid_file = next(cursor)
-            return GridOut(self._collection, file_document=grid_file)
+            return GridOut(
+                self._collection, file_document=grid_file, session=session)
         except StopIteration:
             raise NoFile(
                 "no version %d for filename %r" % (revision, filename))
 
-    def download_to_stream_by_name(self, filename, destination, revision=-1):
+    def download_to_stream_by_name(self, filename, destination, revision=-1,
+                                   session=None):
         """Write the contents of `filename` (with optional `revision`) to
         `destination`.
 
@@ -663,7 +844,7 @@ class GridFSBucket(object):
           my_db = MongoClient().test
           fs = GridFSBucket(my_db)
           # Get file to write to
-          file = open('myfile','w')
+          file = open('myfile','wb')
           fs.download_to_stream_by_name("test_file", file)
 
         Raises :exc:`~gridfs.errors.NoFile` if no such version of
@@ -677,20 +858,27 @@ class GridFSBucket(object):
           - `revision` (optional): Which revision (documents with the same
             filename and different uploadDate) of the file to retrieve.
             Defaults to -1 (the most recent revision).
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
 
         :Note: Revision numbers are defined as follows:
-          0 = the original stored file
-          1 = the first revision
-          2 = the second revision
-          etc...
-          -2 = the second most recent revision
-          -1 = the most recent revision
+
+          - 0 = the original stored file
+          - 1 = the first revision
+          - 2 = the second revision
+          - etc...
+          - -2 = the second most recent revision
+          - -1 = the most recent revision
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
         """
-        gout = self.open_download_stream_by_name(filename, revision)
+        gout = self.open_download_stream_by_name(
+            filename, revision, session=session)
+        for chunk in gout:
+            destination.write(chunk)
 
-        destination.write(gout)
-
-    def rename(self, file_id, new_filename):
+    def rename(self, file_id, new_filename, session=None):
         """Renames the stored file with the specified file_id.
 
         For example::
@@ -706,9 +894,15 @@ class GridFSBucket(object):
         :Parameters:
           - `file_id`: The _id of the file to be renamed.
           - `new_filename`: The new name of the file.
+          - `session` (optional): a
+            :class:`~pymongo.client_session.ClientSession`
+
+        .. versionchanged:: 3.6
+           Added ``session`` parameter.
         """
         result = self._files.update_one({"_id": file_id},
-                                        {"$set": {"filename": new_filename}})
+                                        {"$set": {"filename": new_filename}},
+                                        session=session)
         if not result.matched_count:
             raise NoFile("no files could be renamed %r because none "
                          "matched file_id %i" % (new_filename, file_id))
