@@ -1,4 +1,4 @@
-# Copyright 2011-2015 MongoDB, Inc.
+# Copyright 2011-present MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you
 # may not use this file except in compliance with the License.  You
@@ -30,13 +30,14 @@ except ImportError:
 
 
 from bson import DEFAULT_CODEC_OPTIONS
-from bson.py3compat import imap, itervalues, _unicode
+from bson.py3compat import imap, itervalues, _unicode, integer_types
 from bson.son import SON
 from pymongo import auth, helpers, thread_util, __version__
-from pymongo.common import MAX_MESSAGE_SIZE
+from pymongo.common import MAX_MESSAGE_SIZE, ORDERED_TYPES
 from pymongo.errors import (AutoReconnect,
                             ConnectionFailure,
                             ConfigurationError,
+                            InvalidOperation,
                             DocumentTooLarge,
                             NetworkTimeout,
                             NotMasterError,
@@ -46,7 +47,6 @@ from pymongo.monotonic import time as _time
 from pymongo.network import (command,
                              receive_message,
                              SocketChecker)
-from pymongo.read_concern import DEFAULT_READ_CONCERN
 from pymongo.read_preferences import ReadPreference
 from pymongo.server_type import SERVER_TYPE
 # Always use our backport so we always have support for IP address matching
@@ -111,17 +111,75 @@ except ImportError:
         """Dummy function for platforms that don't provide fcntl."""
         pass
 
+_MAX_TCP_KEEPIDLE = 300
+_MAX_TCP_KEEPINTVL = 10
+_MAX_TCP_KEEPCNT = 9
+
+if sys.platform == 'win32':
+    try:
+        import _winreg as winreg
+    except ImportError:
+        import winreg
+
+    try:
+        with winreg.OpenKey(
+                winreg.HKEY_LOCAL_MACHINE,
+                r"SYSTEM\CurrentControlSet\Services\Tcpip\Parameters") as key:
+            _DEFAULT_TCP_IDLE_MS, _ = winreg.QueryValueEx(key, "KeepAliveTime")
+            _DEFAULT_TCP_INTERVAL_MS, _ = winreg.QueryValueEx(
+                key, "KeepAliveInterval")
+            # Make sure these are integers.
+            if not isinstance(_DEFAULT_TCP_IDLE_MS, integer_types):
+                raise ValueError
+            if not isinstance(_DEFAULT_TCP_INTERVAL_MS, integer_types):
+                raise ValueError
+    except (OSError, ValueError):
+        # We could not check the default values so do not attempt to override.
+        def _set_keepalive_times(dummy):
+            pass
+    else:
+        def _set_keepalive_times(sock):
+            idle_ms = min(_DEFAULT_TCP_IDLE_MS, _MAX_TCP_KEEPIDLE * 1000)
+            interval_ms = min(_DEFAULT_TCP_INTERVAL_MS,
+                              _MAX_TCP_KEEPINTVL * 1000)
+            if (idle_ms < _DEFAULT_TCP_IDLE_MS or
+                    interval_ms < _DEFAULT_TCP_INTERVAL_MS):
+                sock.ioctl(socket.SIO_KEEPALIVE_VALS,
+                           (1, idle_ms, interval_ms))
+else:
+    def _set_tcp_option(sock, tcp_option, max_value):
+        if hasattr(socket, tcp_option):
+            sockopt = getattr(socket, tcp_option)
+            try:
+                # PYTHON-1350 - NetBSD doesn't implement getsockopt for
+                # TCP_KEEPIDLE and friends. Don't attempt to set the
+                # values there.
+                default = sock.getsockopt(socket.IPPROTO_TCP, sockopt)
+                if default > max_value:
+                    sock.setsockopt(socket.IPPROTO_TCP, sockopt, max_value)
+            except socket.error:
+                pass
+
+    def _set_keepalive_times(sock):
+        _set_tcp_option(sock, 'TCP_KEEPIDLE', _MAX_TCP_KEEPIDLE)
+        _set_tcp_option(sock, 'TCP_KEEPINTVL', _MAX_TCP_KEEPINTVL)
+        _set_tcp_option(sock, 'TCP_KEEPCNT', _MAX_TCP_KEEPCNT)
 
 _METADATA = SON([
     ('driver', SON([('name', 'PyMongo'), ('version', __version__)])),
 ])
 
 if sys.platform.startswith('linux'):
+    # platform.linux_distribution was deprecated in Python 3.5.
+    if sys.version_info[:2] < (3, 5):
+        # Distro name and version (e.g. Ubuntu 16.04 xenial)
+        _name = ' '.join([part for part in
+                          platform.linux_distribution() if part])
+    else:
+        _name = platform.system()
     _METADATA['os'] = SON([
         ('type', platform.system()),
-        # Distro name and version (e.g. Ubuntu 16.04 xenial)
-        ('name', ' '.join([part for part in
-                           platform.linux_distribution() if part])),
+        ('name', _name),
         ('architecture', platform.machine()),
         # Kernel version (e.g. 4.4.0-17-generic).
         ('version', platform.release())
@@ -223,7 +281,7 @@ class PoolOptions(object):
                  max_idle_time_ms=None, connect_timeout=None,
                  socket_timeout=None, wait_queue_timeout=None,
                  wait_queue_multiple=None, ssl_context=None,
-                 ssl_match_hostname=True, socket_keepalive=False,
+                 ssl_match_hostname=True, socket_keepalive=True,
                  event_listeners=None, appname=None):
 
         self.__max_pool_size = max_pool_size
@@ -350,7 +408,7 @@ class SocketInfo(object):
         self.address = address
         self.authset = set()
         self.closed = False
-        self.last_checkout = _time()
+        self.last_checkin = _time()
         self.is_writable = ismaster.is_writable if ismaster else None
         self.max_wire_version = ismaster.max_wire_version if ismaster else None
         self.max_bson_size = ismaster.max_bson_size if ismaster else None
@@ -358,6 +416,8 @@ class SocketInfo(object):
             ismaster.max_message_size if ismaster else MAX_MESSAGE_SIZE)
         self.max_write_batch_size = (
             ismaster.max_write_batch_size if ismaster else None)
+        self.supports_sessions = (
+            ismaster and ismaster.logical_session_timeout_minutes is not None)
 
         self.listeners = pool.opts.event_listeners
 
@@ -374,11 +434,14 @@ class SocketInfo(object):
                 read_preference=ReadPreference.PRIMARY,
                 codec_options=DEFAULT_CODEC_OPTIONS, check=True,
                 allowable_errors=None, check_keys=False,
-                read_concern=DEFAULT_READ_CONCERN,
+                read_concern=None,
                 write_concern=None,
                 parse_write_concern_error=False,
-                collation=None):
-        """Execute a command or raise ConnectionFailure or OperationFailure.
+                collation=None,
+                session=None,
+                client=None,
+                retryable_write=False):
+        """Execute a command or raise an error.
 
         :Parameters:
           - `dbname`: name of the database on which to run the command
@@ -394,8 +457,13 @@ class SocketInfo(object):
           - `parse_write_concern_error`: Whether to parse the
             ``writeConcernError`` field in the command response.
           - `collation`: The collation for this command.
+          - `session`: optional ClientSession instance.
+          - `client`: optional MongoClient for gossipping $clusterTime.
+          - `retryable_write`: True if this command is a retryable write.
         """
-        if self.max_wire_version < 4 and not read_concern.ok_for_legacy:
+        self.validate_session(client, session)
+        if (read_concern and self.max_wire_version < 4
+                and not read_concern.ok_for_legacy):
             raise ConfigurationError(
                 'read concern level of %s is not valid '
                 'with a max wire version of %d.'
@@ -409,12 +477,21 @@ class SocketInfo(object):
         elif self.max_wire_version < 5 and collation is not None:
             raise ConfigurationError(
                 'Must be connected to MongoDB 3.4+ to use a collation.')
+
+        if (client or session) and not isinstance(spec, ORDERED_TYPES):
+            # Ensure command name remains in first place.
+            spec = SON(spec)
+        if session:
+            spec['lsid'] = session._use_lsid()
+            if retryable_write:
+                spec['txnNumber'] = session._transaction_id()
+        self.send_cluster_time(spec, session, client)
         try:
             return command(self.sock, dbname, spec, slave_ok,
                            self.is_mongos, read_preference, codec_options,
-                           check, allowable_errors, self.address,
-                           check_keys, self.listeners, self.max_bson_size,
-                           read_concern,
+                           session, client, check, allowable_errors,
+                           self.address, check_keys, self.listeners,
+                           self.max_bson_size, read_concern,
                            parse_write_concern_error=parse_write_concern_error,
                            collation=collation)
         except OperationFailure:
@@ -440,14 +517,14 @@ class SocketInfo(object):
         except BaseException as error:
             self._raise_connection_failure(error)
 
-    def receive_message(self, operation, request_id):
+    def receive_message(self, request_id):
         """Receive a raw BSON message or raise ConnectionFailure.
 
         If any exception is raised, the socket is closed.
         """
         try:
-            return receive_message(
-                self.sock, operation, request_id, self.max_message_size)
+            return receive_message(self.sock, request_id,
+                                   self.max_message_size)
         except BaseException as error:
             self._raise_connection_failure(error)
 
@@ -469,8 +546,8 @@ class SocketInfo(object):
 
         self.send_message(msg, max_doc_size)
         if with_last_error:
-            response = self.receive_message(1, request_id)
-            return helpers._check_gle_response(response)
+            reply = self.receive_message(request_id)
+            return helpers._check_gle_response(reply.command_response())
 
     def write_command(self, request_id, msg):
         """Send "insert" etc. command, returning response as a dict.
@@ -482,9 +559,8 @@ class SocketInfo(object):
           - `msg`: bytes, the command message.
         """
         self.send_message(msg, 0)
-        response = helpers._unpack_response(self.receive_message(1, request_id))
-        assert response['number_returned'] == 1
-        result = response['data'][0]
+        reply = self.receive_message(request_id)
+        result = reply.command_response()
 
         # Raises NotMasterError or OperationFailure.
         helpers._check_command_response(result)
@@ -523,14 +599,35 @@ class SocketInfo(object):
         auth.authenticate(credentials, self)
         self.authset.add(credentials)
 
+    def validate_session(self, client, session):
+        """Validate this session before use with client.
+
+        Raises error if this session is logged in as a different user or
+        the client is not the one that created the session.
+        """
+        if session:
+            if session._client is not client:
+                raise InvalidOperation(
+                    'Can only use session with the MongoClient that'
+                    ' started it')
+            if session._authset != self.authset:
+                raise InvalidOperation(
+                    'Cannot use session after authenticating with different'
+                    ' credentials')
+
     def close(self):
         self.closed = True
         # Avoid exceptions on interpreter shutdown.
         try:
             self.sock.close()
-        except:
+        except Exception:
             pass
 
+    def send_cluster_time(self, command, session, client):
+        """Add cluster time for MongoDB >= 3.6."""
+        if self.max_wire_version >= 6 and client:
+            client._send_cluster_time(command, session)
+
     def _raise_connection_failure(self, error):
         # Catch *all* exceptions from socket methods and close the socket. In
         # regular Python, socket operations only raise socket.error, even if
@@ -619,6 +716,8 @@ def _create_connection(address, options):
             sock.settimeout(options.connect_timeout)
             sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE,
                             options.socket_keepalive)
+            if options.socket_keepalive:
+                _set_keepalive_times(sock)
             sock.connect(sa)
             return sock
         except socket.error as e:
@@ -716,11 +815,11 @@ class Pool:
             sock_info.close()
 
     def remove_stale_sockets(self):
-        with self.lock:
-            if self.opts.max_idle_time_ms is not None:
+        if self.opts.max_idle_time_ms is not None:
+            with self.lock:
                 for sock_info in self.sockets.copy():
-                    age = _time() - sock_info.last_checkout
-                    if age > self.opts.max_idle_time_ms:
+                    idle_time_ms = 1000 * _time() - sock_info.last_checkin
+                    if idle_time_ms > self.opts.max_idle_time_ms:
                         self.sockets.remove(sock_info)
                         sock_info.close()
 
@@ -753,7 +852,9 @@ class Pool:
                             False,
                             False,
                             ReadPreference.PRIMARY,
-                            DEFAULT_CODEC_OPTIONS))
+                            DEFAULT_CODEC_OPTIONS,
+                            None,
+                            None))
             else:
                 ismaster = None
             return SocketInfo(sock, self, ismaster, self.address)
@@ -820,28 +921,20 @@ class Pool:
                 # set.pop() isn't atomic in Jython less than 2.7, see
                 # http://bugs.jython.org/issue1854
                 with self.lock:
-                    sock_info, from_pool = self.sockets.pop(), True
+                    # Can raise ConnectionFailure.
+                    sock_info = self.sockets.pop()
             except KeyError:
                 # Can raise ConnectionFailure or CertificateError.
-                sock_info, from_pool = self.connect(), False
-            # If socket is idle, open a new one.
-            if self.opts.max_idle_time_ms is not None:
-                age = _time() - sock_info.last_checkout
-                if age > self.opts.max_idle_time_ms:
-                    sock_info.close()
-                    sock_info, from_pool = self.connect(), False
-
-            if from_pool:
+                sock_info = self.connect()
+            else:
                 # Can raise ConnectionFailure.
                 sock_info = self._check(sock_info)
-
         except:
             self._socket_semaphore.release()
             with self.lock:
                 self.active_sockets -= 1
             raise
 
-        sock_info.last_checkout = _time()
         return sock_info
 
     def return_socket(self, sock_info):
@@ -852,6 +945,7 @@ class Pool:
             if sock_info.pool_id != self.pool_id:
                 sock_info.close()
             elif not sock_info.closed:
+                sock_info.last_checkin = _time()
                 with self.lock:
                     self.sockets.add(sock_info)
 
@@ -860,34 +954,35 @@ class Pool:
             self.active_sockets -= 1
 
     def _check(self, sock_info):
-        """This side-effecty function checks if this pool has been reset since
-        the last time this socket was used, or if the socket has been closed by
-        some external network error, and if so, attempts to create a new socket.
-        If this connection attempt fails we reset the pool and reraise the
+        """This side-effecty function checks if this socket has been idle for
+        for longer than the max idle time, or if the socket has been closed by
+        some external network error, and if so, attempts to create a new
+        socket. If this connection attempt fails we raise the
         ConnectionFailure.
 
         Checking sockets lets us avoid seeing *some*
         :class:`~pymongo.errors.AutoReconnect` exceptions on server
-        hiccups, etc. We only do this if it's been > 1 second since
-        the last socket checkout, to keep performance reasonable - we
-        can't avoid AutoReconnects completely anyway.
+        hiccups, etc. We only check if the socket was closed by an external
+        error if it has been > 1 second since the socket was checked into the
+        pool, to keep performance reasonable - we can't avoid AutoReconnects
+        completely anyway.
         """
-        error = False
+        # How long since socket was last checked in.
+        idle_time_seconds = _time() - sock_info.last_checkin
+        # If socket is idle, open a new one.
+        if (self.opts.max_idle_time_ms is not None and
+                idle_time_seconds * 1000 > self.opts.max_idle_time_ms):
+            sock_info.close()
+            return self.connect()
 
-        # How long since socket was last checked out.
-        age = _time() - sock_info.last_checkout
-        if (self._check_interval_seconds is not None
-                and (
-                    0 == self._check_interval_seconds
-                    or age > self._check_interval_seconds)):
+        if (self._check_interval_seconds is not None and (
+                0 == self._check_interval_seconds or
+                idle_time_seconds > self._check_interval_seconds)):
             if self.socket_checker.socket_closed(sock_info.sock):
                 sock_info.close()
-                error = True
+                return self.connect()
 
-        if not error:
-            return sock_info
-        else:
-            return self.connect()
+        return sock_info
 
     def _raise_wait_queue_timeout(self):
         raise ConnectionFailure(
